@@ -5,18 +5,23 @@ use futures::future::Future;
 use hyper::header;
 use hyper::server::conn::AddrStream;
 use hyper::service::{make_service_fn, service_fn};
-use hyper::{Body, Request, Response, Server};
+use hyper::{Body, Method, Request, Response, Server};
 
+use garage_util::data::*;
 use garage_util::error::Error as GarageError;
 
 use garage_model::garage::Garage;
+use garage_model::key_table::Key;
+
+use garage_table::util::*;
 
 use crate::error::*;
-use crate::signature::check_signature;
+use crate::signature::payload::check_payload_signature;
 
 use crate::helpers::*;
 use crate::s3_bucket::*;
 use crate::s3_copy::*;
+use crate::s3_cors::*;
 use crate::s3_delete::*;
 use crate::s3_get::*;
 use crate::s3_list::*;
@@ -87,7 +92,10 @@ async fn handler(
 }
 
 async fn handler_inner(garage: Arc<Garage>, req: Request<Body>) -> Result<Response<Body>, Error> {
-	let (api_key, content_sha256) = check_signature(&garage, &req).await?;
+	let (api_key, content_sha256) = check_payload_signature(&garage, &req).await?;
+	let api_key = api_key.ok_or_else(|| {
+		Error::Forbidden("Garage does not support anonymous access yet".to_string())
+	})?;
 
 	let authority = req
 		.headers()
@@ -97,18 +105,39 @@ async fn handler_inner(garage: Arc<Garage>, req: Request<Body>) -> Result<Respon
 
 	let host = authority_to_host(authority)?;
 
-	let bucket = garage
+	let bucket_name = garage
 		.config
 		.s3_api
 		.root_domain
 		.as_ref()
 		.and_then(|root_domain| host_to_bucket(&host, root_domain));
 
-	let endpoint = Endpoint::from_request(&req, bucket.map(ToOwned::to_owned))?;
+	let (endpoint, bucket_name) = Endpoint::from_request(&req, bucket_name.map(ToOwned::to_owned))?;
+	debug!("Endpoint: {:?}", endpoint);
+
+	let bucket_name = match bucket_name {
+		None => return handle_request_without_bucket(garage, req, api_key, endpoint).await,
+		Some(bucket) => bucket.to_string(),
+	};
+
+	// Special code path for CreateBucket API endpoint
+	if let Endpoint::CreateBucket {} = endpoint {
+		return handle_create_bucket(&garage, req, content_sha256, api_key, bucket_name).await;
+	}
+
+	let bucket_id = resolve_bucket(&garage, &bucket_name, &api_key).await?;
+	let bucket = garage
+		.bucket_table
+		.get(&EmptyKey, &bucket_id)
+		.await?
+		.filter(|b| !b.state.is_deleted())
+		.ok_or(Error::NoSuchBucket)?;
+
 	let allowed = match endpoint.authorization_type() {
-		Authorization::None => true,
-		Authorization::Read(bucket) => api_key.allow_read(bucket),
-		Authorization::Write(bucket) => api_key.allow_write(bucket),
+		Authorization::Read => api_key.allow_read(&bucket_id),
+		Authorization::Write => api_key.allow_write(&bucket_id),
+		Authorization::Owner => api_key.allow_owner(&bucket_id),
+		_ => unreachable!(),
 	};
 
 	if !allowed {
@@ -117,12 +146,24 @@ async fn handler_inner(garage: Arc<Garage>, req: Request<Body>) -> Result<Respon
 		));
 	}
 
-	match endpoint {
-		Endpoint::ListBuckets => handle_list_buckets(&api_key),
-		Endpoint::HeadObject { bucket, key, .. } => handle_head(garage, &req, &bucket, &key).await,
-		Endpoint::GetObject { bucket, key, .. } => handle_get(garage, &req, &bucket, &key).await,
+	// Look up what CORS rule might apply to response.
+	// Requests for methods different than GET, HEAD or POST
+	// are always preflighted, i.e. the browser should make
+	// an OPTIONS call before to check it is allowed
+	let matching_cors_rule = match *req.method() {
+		Method::GET | Method::HEAD | Method::POST => find_matching_cors_rule(&bucket, &req)?,
+		_ => None,
+	};
+
+	let resp = match endpoint {
+		Endpoint::Options => handle_options(&req, &bucket).await,
+		Endpoint::HeadObject {
+			key, part_number, ..
+		} => handle_head(garage, &req, bucket_id, &key, part_number).await,
+		Endpoint::GetObject {
+			key, part_number, ..
+		} => handle_get(garage, &req, bucket_id, &key, part_number).await,
 		Endpoint::UploadPart {
-			bucket,
 			key,
 			part_number,
 			upload_id,
@@ -130,7 +171,7 @@ async fn handler_inner(garage: Arc<Garage>, req: Request<Body>) -> Result<Respon
 			handle_put_part(
 				garage,
 				req,
-				&bucket,
+				bucket_id,
 				&key,
 				part_number,
 				&upload_id,
@@ -138,64 +179,57 @@ async fn handler_inner(garage: Arc<Garage>, req: Request<Body>) -> Result<Respon
 			)
 			.await
 		}
-		Endpoint::CopyObject { bucket, key } => {
-			let copy_source = req.headers().get("x-amz-copy-source").unwrap().to_str()?;
-			let copy_source = percent_encoding::percent_decode_str(copy_source).decode_utf8()?;
-			let (source_bucket, source_key) = parse_bucket_key(&copy_source, None)?;
-			if !api_key.allow_read(source_bucket) {
-				return Err(Error::Forbidden(format!(
-					"Reading from bucket {} not allowed for this key",
-					source_bucket
-				)));
-			}
-			let source_key = source_key.ok_or_bad_request("No source key specified")?;
-			handle_copy(garage, &req, &bucket, &key, source_bucket, source_key).await
-		}
-		Endpoint::PutObject { bucket, key } => {
-			handle_put(garage, req, &bucket, &key, content_sha256).await
-		}
-		Endpoint::AbortMultipartUpload {
-			bucket,
-			key,
-			upload_id,
-		} => handle_abort_multipart_upload(garage, &bucket, &key, &upload_id).await,
-		Endpoint::DeleteObject { bucket, key, .. } => handle_delete(garage, &bucket, &key).await,
-		Endpoint::CreateMultipartUpload { bucket, key } => {
-			handle_create_multipart_upload(garage, &req, &bucket, &key).await
-		}
-		Endpoint::CompleteMultipartUpload {
-			bucket,
+		Endpoint::CopyObject { key } => handle_copy(garage, &api_key, &req, bucket_id, &key).await,
+		Endpoint::UploadPartCopy {
 			key,
+			part_number,
 			upload_id,
 		} => {
-			handle_complete_multipart_upload(garage, req, &bucket, &key, &upload_id, content_sha256)
-				.await
+			handle_upload_part_copy(
+				garage,
+				&api_key,
+				&req,
+				bucket_id,
+				&key,
+				part_number,
+				&upload_id,
+			)
+			.await
 		}
-		Endpoint::CreateBucket { bucket } => {
-			debug!(
-				"Body: {}",
-				std::str::from_utf8(&hyper::body::to_bytes(req.into_body()).await?)
-					.unwrap_or("<invalid utf8>")
-			);
-			let empty_body: Body = Body::from(vec![]);
-			let response = Response::builder()
-				.header("Location", format!("/{}", bucket))
-				.body(empty_body)
-				.unwrap();
-			Ok(response)
+		Endpoint::PutObject { key } => {
+			handle_put(garage, req, bucket_id, &key, &api_key, content_sha256).await
 		}
-		Endpoint::HeadBucket { .. } => {
+		Endpoint::AbortMultipartUpload { key, upload_id } => {
+			handle_abort_multipart_upload(garage, bucket_id, &key, &upload_id).await
+		}
+		Endpoint::DeleteObject { key, .. } => handle_delete(garage, bucket_id, &key).await,
+		Endpoint::CreateMultipartUpload { key } => {
+			handle_create_multipart_upload(garage, &req, &bucket_name, bucket_id, &key).await
+		}
+		Endpoint::CompleteMultipartUpload { key, upload_id } => {
+			handle_complete_multipart_upload(
+				garage,
+				req,
+				&bucket_name,
+				bucket_id,
+				&key,
+				&upload_id,
+				content_sha256,
+			)
+			.await
+		}
+		Endpoint::CreateBucket {} => unreachable!(),
+		Endpoint::HeadBucket {} => {
 			let empty_body: Body = Body::from(vec![]);
 			let response = Response::builder().body(empty_body).unwrap();
 			Ok(response)
 		}
-		Endpoint::DeleteBucket { .. } => Err(Error::Forbidden(
-			"Cannot delete buckets using S3 api, please talk to Garage directly".into(),
-		)),
-		Endpoint::GetBucketLocation { .. } => handle_get_bucket_location(garage),
-		Endpoint::GetBucketVersioning { .. } => handle_get_bucket_versioning(),
+		Endpoint::DeleteBucket {} => {
+			handle_delete_bucket(&garage, bucket_id, bucket_name, api_key).await
+		}
+		Endpoint::GetBucketLocation {} => handle_get_bucket_location(garage),
+		Endpoint::GetBucketVersioning {} => handle_get_bucket_versioning(),
 		Endpoint::ListObjects {
-			bucket,
 			delimiter,
 			encoding_type,
 			marker,
@@ -205,21 +239,23 @@ async fn handler_inner(garage: Arc<Garage>, req: Request<Body>) -> Result<Respon
 			handle_list(
 				garage,
 				&ListObjectsQuery {
+					common: ListQueryCommon {
+						bucket_name,
+						bucket_id,
+						delimiter: delimiter.map(|d| d.to_string()),
+						page_size: max_keys.map(|p| p.clamp(1, 1000)).unwrap_or(1000),
+						prefix: prefix.unwrap_or_default(),
+						urlencode_resp: encoding_type.map(|e| e == "url").unwrap_or(false),
+					},
 					is_v2: false,
-					bucket,
-					delimiter: delimiter.map(|d| d.to_string()),
-					max_keys: max_keys.unwrap_or(1000),
-					prefix: prefix.unwrap_or_default(),
 					marker,
 					continuation_token: None,
 					start_after: None,
-					urlencode_resp: encoding_type.map(|e| e == "url").unwrap_or(false),
 				},
 			)
 			.await
 		}
 		Endpoint::ListObjectsV2 {
-			bucket,
 			delimiter,
 			encoding_type,
 			max_keys,
@@ -233,15 +269,18 @@ async fn handler_inner(garage: Arc<Garage>, req: Request<Body>) -> Result<Respon
 				handle_list(
 					garage,
 					&ListObjectsQuery {
+						common: ListQueryCommon {
+							bucket_name,
+							bucket_id,
+							delimiter: delimiter.map(|d| d.to_string()),
+							page_size: max_keys.map(|p| p.clamp(1, 1000)).unwrap_or(1000),
+							urlencode_resp: encoding_type.map(|e| e == "url").unwrap_or(false),
+							prefix: prefix.unwrap_or_default(),
+						},
 						is_v2: true,
-						bucket,
-						delimiter: delimiter.map(|d| d.to_string()),
-						max_keys: max_keys.unwrap_or(1000),
-						prefix: prefix.unwrap_or_default(),
 						marker: None,
 						continuation_token,
 						start_after,
-						urlencode_resp: encoding_type.map(|e| e == "url").unwrap_or(false),
 					},
 				)
 				.await
@@ -252,14 +291,106 @@ async fn handler_inner(garage: Arc<Garage>, req: Request<Body>) -> Result<Respon
 				)))
 			}
 		}
-		Endpoint::DeleteObjects { bucket } => {
-			handle_delete_objects(garage, &bucket, req, content_sha256).await
+		Endpoint::ListMultipartUploads {
+			delimiter,
+			encoding_type,
+			key_marker,
+			max_uploads,
+			prefix,
+			upload_id_marker,
+		} => {
+			handle_list_multipart_upload(
+				garage,
+				&ListMultipartUploadsQuery {
+					common: ListQueryCommon {
+						bucket_name,
+						bucket_id,
+						delimiter: delimiter.map(|d| d.to_string()),
+						page_size: max_uploads.map(|p| p.clamp(1, 1000)).unwrap_or(1000),
+						prefix: prefix.unwrap_or_default(),
+						urlencode_resp: encoding_type.map(|e| e == "url").unwrap_or(false),
+					},
+					key_marker,
+					upload_id_marker,
+				},
+			)
+			.await
 		}
-		Endpoint::PutBucketWebsite { bucket } => {
-			handle_put_website(garage, bucket, req, content_sha256).await
+		Endpoint::ListParts {
+			key,
+			max_parts,
+			part_number_marker,
+			upload_id,
+		} => {
+			handle_list_parts(
+				garage,
+				&ListPartsQuery {
+					bucket_name,
+					bucket_id,
+					key,
+					upload_id,
+					part_number_marker: part_number_marker.map(|p| p.clamp(1, 10000)),
+					max_parts: max_parts.map(|p| p.clamp(1, 1000)).unwrap_or(1000),
+				},
+			)
+			.await
 		}
-		Endpoint::DeleteBucketWebsite { bucket } => handle_delete_website(garage, bucket).await,
+		Endpoint::DeleteObjects {} => {
+			handle_delete_objects(garage, bucket_id, req, content_sha256).await
+		}
+		Endpoint::GetBucketWebsite {} => handle_get_website(&bucket).await,
+		Endpoint::PutBucketWebsite {} => {
+			handle_put_website(garage, bucket_id, req, content_sha256).await
+		}
+		Endpoint::DeleteBucketWebsite {} => handle_delete_website(garage, bucket_id).await,
+		Endpoint::GetBucketCors {} => handle_get_cors(&bucket).await,
+		Endpoint::PutBucketCors {} => handle_put_cors(garage, bucket_id, req, content_sha256).await,
+		Endpoint::DeleteBucketCors {} => handle_delete_cors(garage, bucket_id).await,
 		endpoint => Err(Error::NotImplemented(endpoint.name().to_owned())),
+	};
+
+	// If request was a success and we have a CORS rule that applies to it,
+	// add the corresponding CORS headers to the response
+	let mut resp_ok = resp?;
+	if let Some(rule) = matching_cors_rule {
+		add_cors_headers(&mut resp_ok, rule)
+			.ok_or_internal_error("Invalid bucket CORS configuration")?;
+	}
+
+	Ok(resp_ok)
+}
+
+async fn handle_request_without_bucket(
+	garage: Arc<Garage>,
+	_req: Request<Body>,
+	api_key: Key,
+	endpoint: Endpoint,
+) -> Result<Response<Body>, Error> {
+	match endpoint {
+		Endpoint::ListBuckets => handle_list_buckets(&garage, &api_key).await,
+		endpoint => Err(Error::NotImplemented(endpoint.name().to_owned())),
+	}
+}
+
+#[allow(clippy::ptr_arg)]
+pub async fn resolve_bucket(
+	garage: &Garage,
+	bucket_name: &String,
+	api_key: &Key,
+) -> Result<Uuid, Error> {
+	let api_key_params = api_key
+		.state
+		.as_option()
+		.ok_or_internal_error("Key should not be deleted at this point")?;
+
+	if let Some(Some(bucket_id)) = api_key_params.local_aliases.get(bucket_name) {
+		Ok(*bucket_id)
+	} else {
+		Ok(garage
+			.bucket_helper()
+			.resolve_global_bucket_name(bucket_name)
+			.await?
+			.ok_or(Error::NoSuchBucket)?)
 	}
 }
 
@@ -268,7 +399,7 @@ async fn handler_inner(garage: Arc<Garage>, req: Request<Body>) -> Result<Respon
 ///
 /// S3 internally manages only buckets and keys. This function splits
 /// an HTTP path to get the corresponding bucket name and key.
-fn parse_bucket_key<'a>(
+pub fn parse_bucket_key<'a>(
 	path: &'a str,
 	host_bucket: Option<&'a str>,
 ) -> Result<(&'a str, Option<&'a str>), Error> {
