@@ -2,7 +2,7 @@
 #[macro_use]
 extern crate log;
 
-mod error;
+pub mod error;
 pub use error::Error;
 
 mod encoding;
@@ -15,6 +15,7 @@ mod signature;
 pub mod helpers;
 mod s3_bucket;
 mod s3_copy;
+pub mod s3_cors;
 mod s3_delete;
 pub mod s3_get;
 mod s3_list;
