@@ -16,6 +16,12 @@ pub fn xmlns_tag<S: Serializer>(_v: &(), s: S) -> Result<S::Ok, S::Error> {
 #[derive(Debug, Serialize, Deserialize, PartialEq, Eq, PartialOrd, Ord)]
 pub struct Value(#[serde(rename = "$value")] pub String);
 
+impl From<&str> for Value {
+	fn from(s: &str) -> Value {
+		Value(s.to_string())
+	}
+}
+
 #[derive(Debug, Serialize, Deserialize, PartialEq, Eq, PartialOrd, Ord)]
 pub struct IntValue(#[serde(rename = "$value")] pub i64);
 
@@ -27,12 +33,6 @@ pub struct Bucket {
 	pub name: Value,
 }
 
-#[derive(Debug, Serialize, PartialEq)]
-pub struct DisplayName(#[serde(rename = "$value")] pub String);
-
-#[derive(Debug, Serialize, PartialEq)]
-pub struct Id(#[serde(rename = "$value")] pub String);
-
 #[derive(Debug, Serialize, PartialEq)]
 pub struct Owner {
 	#[serde(rename = "DisplayName")]
@@ -107,14 +107,6 @@ pub struct DeleteResult {
 	pub errors: Vec<DeleteError>,
 }
 
-#[derive(Debug, Serialize, PartialEq)]
-pub struct CopyObjectResult {
-	#[serde(rename = "LastModified")]
-	pub last_modified: Value,
-	#[serde(rename = "ETag")]
-	pub etag: Value,
-}
-
 #[derive(Debug, Serialize, PartialEq)]
 pub struct InitiateMultipartUploadResult {
 	#[serde(serialize_with = "xmlns_tag")]
@@ -141,6 +133,100 @@ pub struct CompleteMultipartUploadResult {
 	pub etag: Value,
 }
 
+#[derive(Debug, Serialize, PartialEq)]
+pub struct Initiator {
+	#[serde(rename = "DisplayName")]
+	pub display_name: Value,
+	#[serde(rename = "ID")]
+	pub id: Value,
+}
+
+#[derive(Debug, Serialize, PartialEq)]
+pub struct ListMultipartItem {
+	#[serde(rename = "Initiated")]
+	pub initiated: Value,
+	#[serde(rename = "Initiator")]
+	pub initiator: Initiator,
+	#[serde(rename = "Key")]
+	pub key: Value,
+	#[serde(rename = "UploadId")]
+	pub upload_id: Value,
+	#[serde(rename = "Owner")]
+	pub owner: Owner,
+	#[serde(rename = "StorageClass")]
+	pub storage_class: Value,
+}
+
+#[derive(Debug, Serialize, PartialEq)]
+pub struct ListMultipartUploadsResult {
+	#[serde(serialize_with = "xmlns_tag")]
+	pub xmlns: (),
+	#[serde(rename = "Bucket")]
+	pub bucket: Value,
+	#[serde(rename = "KeyMarker")]
+	pub key_marker: Option<Value>,
+	#[serde(rename = "UploadIdMarker")]
+	pub upload_id_marker: Option<Value>,
+	#[serde(rename = "NextKeyMarker")]
+	pub next_key_marker: Option<Value>,
+	#[serde(rename = "NextUploadIdMarker")]
+	pub next_upload_id_marker: Option<Value>,
+	#[serde(rename = "Prefix")]
+	pub prefix: Value,
+	#[serde(rename = "Delimiter")]
+	pub delimiter: Option<Value>,
+	#[serde(rename = "MaxUploads")]
+	pub max_uploads: IntValue,
+	#[serde(rename = "IsTruncated")]
+	pub is_truncated: Value,
+	#[serde(rename = "Upload")]
+	pub upload: Vec<ListMultipartItem>,
+	#[serde(rename = "CommonPrefixes")]
+	pub common_prefixes: Vec<CommonPrefix>,
+	#[serde(rename = "EncodingType")]
+	pub encoding_type: Option<Value>,
+}
+
+#[derive(Debug, Serialize, PartialEq)]
+pub struct PartItem {
+	#[serde(rename = "ETag")]
+	pub etag: Value,
+	#[serde(rename = "LastModified")]
+	pub last_modified: Value,
+	#[serde(rename = "PartNumber")]
+	pub part_number: IntValue,
+	#[serde(rename = "Size")]
+	pub size: IntValue,
+}
+
+#[derive(Debug, Serialize, PartialEq)]
+pub struct ListPartsResult {
+	#[serde(serialize_with = "xmlns_tag")]
+	pub xmlns: (),
+	#[serde(rename = "Bucket")]
+	pub bucket: Value,
+	#[serde(rename = "Key")]
+	pub key: Value,
+	#[serde(rename = "UploadId")]
+	pub upload_id: Value,
+	#[serde(rename = "PartNumberMarker")]
+	pub part_number_marker: Option<IntValue>,
+	#[serde(rename = "NextPartNumberMarker")]
+	pub next_part_number_marker: Option<IntValue>,
+	#[serde(rename = "MaxParts")]
+	pub max_parts: IntValue,
+	#[serde(rename = "IsTruncated")]
+	pub is_truncated: Value,
+	#[serde(rename = "Part", default)]
+	pub parts: Vec<PartItem>,
+	#[serde(rename = "Initiator")]
+	pub initiator: Initiator,
+	#[serde(rename = "Owner")]
+	pub owner: Owner,
+	#[serde(rename = "StorageClass")]
+	pub storage_class: Value,
+}
+
 #[derive(Debug, Serialize, PartialEq)]
 pub struct ListBucketItem {
 	#[serde(rename = "Key")]
@@ -372,24 +458,6 @@ mod tests {
 		Ok(())
 	}
 
-	#[test]
-	fn copy_object_result() -> Result<(), ApiError> {
-		let copy_result = CopyObjectResult {
-			last_modified: Value(msec_to_rfc3339(0)),
-			etag: Value("9b2cf535f27731c974343645a3985328".to_string()),
-		};
-		assert_eq!(
-			to_xml_with_header(&copy_result)?,
-			"<?xml version=\"1.0\" encoding=\"UTF-8\"?>\
-<CopyObjectResult>\
-    <LastModified>1970-01-01T00:00:00.000Z</LastModified>\
-    <ETag>9b2cf535f27731c974343645a3985328</ETag>\
-</CopyObjectResult>\
-			"
-		);
-		Ok(())
-	}
-
 	#[test]
 	fn initiate_multipart_upload_result() -> Result<(), ApiError> {
 		let result = InitiateMultipartUploadResult {
@@ -417,7 +485,7 @@ mod tests {
 			location: Some(Value("https://garage.tld/mybucket/a/plop".to_string())),
 			bucket: Value("mybucket".to_string()),
 			key: Value("a/plop".to_string()),
-			etag: Value("3858f62230ac3c915f300c664312c11f-9".to_string()),
+			etag: Value("\"3858f62230ac3c915f300c664312c11f-9\"".to_string()),
 		};
 		assert_eq!(
 			to_xml_with_header(&result)?,
@@ -426,12 +494,64 @@ mod tests {
 	<Location>https://garage.tld/mybucket/a/plop</Location>\
 	<Bucket>mybucket</Bucket>\
 	<Key>a/plop</Key>\
-	<ETag>3858f62230ac3c915f300c664312c11f-9</ETag>\
+	<ETag>&quot;3858f62230ac3c915f300c664312c11f-9&quot;</ETag>\
 </CompleteMultipartUploadResult>"
 		);
 		Ok(())
 	}
 
+	#[test]
+	fn list_multipart_uploads_result() -> Result<(), ApiError> {
+		let result = ListMultipartUploadsResult {
+			xmlns: (),
+			bucket: Value("example-bucket".to_string()),
+			key_marker: None,
+			next_key_marker: None,
+			upload_id_marker: None,
+			encoding_type: None,
+			next_upload_id_marker: None,
+			upload: vec![],
+			delimiter: Some(Value("/".to_string())),
+			prefix: Value("photos/2006/".to_string()),
+			max_uploads: IntValue(1000),
+			is_truncated: Value("false".to_string()),
+			common_prefixes: vec![
+				CommonPrefix {
+					prefix: Value("photos/2006/February/".to_string()),
+				},
+				CommonPrefix {
+					prefix: Value("photos/2006/January/".to_string()),
+				},
+				CommonPrefix {
+					prefix: Value("photos/2006/March/".to_string()),
+				},
+			],
+		};
+
+		assert_eq!(
+			to_xml_with_header(&result)?,
+			"<?xml version=\"1.0\" encoding=\"UTF-8\"?>\
+<ListMultipartUploadsResult xmlns=\"http://s3.amazonaws.com/doc/2006-03-01/\">\
+	<Bucket>example-bucket</Bucket>\
+	<Prefix>photos/2006/</Prefix>\
+	<Delimiter>/</Delimiter>\
+	<MaxUploads>1000</MaxUploads>\
+	<IsTruncated>false</IsTruncated>\
+	<CommonPrefixes>\
+		<Prefix>photos/2006/February/</Prefix>\
+	</CommonPrefixes>\
+	<CommonPrefixes>\
+		<Prefix>photos/2006/January/</Prefix>\
+	</CommonPrefixes>\
+	<CommonPrefixes>\
+		<Prefix>photos/2006/March/</Prefix>\
+	</CommonPrefixes>\
+</ListMultipartUploadsResult>"
+		);
+
+		Ok(())
+	}
+
 	#[test]
 	fn list_objects_v1_1() -> Result<(), ApiError> {
 		let result = ListBucketResult {
@@ -451,7 +571,7 @@ mod tests {
 			contents: vec![ListBucketItem {
 				key: Value("sample.jpg".to_string()),
 				last_modified: Value(msec_to_rfc3339(0)),
-				etag: Value("bf1d737a4d46a19f3bced6905cc8b902".to_string()),
+				etag: Value("\"bf1d737a4d46a19f3bced6905cc8b902\"".to_string()),
 				size: IntValue(142863),
 				storage_class: Value("STANDARD".to_string()),
 			}],
@@ -472,7 +592,7 @@ mod tests {
   <Contents>\
     <Key>sample.jpg</Key>\
     <LastModified>1970-01-01T00:00:00.000Z</LastModified>\
-    <ETag>bf1d737a4d46a19f3bced6905cc8b902</ETag>\
+    <ETag>&quot;bf1d737a4d46a19f3bced6905cc8b902&quot;</ETag>\
     <Size>142863</Size>\
     <StorageClass>STANDARD</StorageClass>\
   </Contents>\
@@ -550,7 +670,7 @@ mod tests {
 			contents: vec![ListBucketItem {
 				key: Value("ExampleObject.txt".to_string()),
 				last_modified: Value(msec_to_rfc3339(0)),
-				etag: Value("599bab3ed2c697f1d26842727561fd94".to_string()),
+				etag: Value("\"599bab3ed2c697f1d26842727561fd94\"".to_string()),
 				size: IntValue(857),
 				storage_class: Value("REDUCED_REDUNDANCY".to_string()),
 			}],
@@ -568,7 +688,7 @@ mod tests {
   <Contents>\
     <Key>ExampleObject.txt</Key>\
     <LastModified>1970-01-01T00:00:00.000Z</LastModified>\
-    <ETag>599bab3ed2c697f1d26842727561fd94</ETag>\
+    <ETag>&quot;599bab3ed2c697f1d26842727561fd94&quot;</ETag>\
     <Size>857</Size>\
     <StorageClass>REDUCED_REDUNDANCY</StorageClass>\
   </Contents>\
@@ -598,7 +718,7 @@ mod tests {
 			contents: vec![ListBucketItem {
 				key: Value("happyfacex.jpg".to_string()),
 				last_modified: Value(msec_to_rfc3339(0)),
-				etag: Value("70ee1738b6b21e2c8a43f3a5ab0eee71".to_string()),
+				etag: Value("\"70ee1738b6b21e2c8a43f3a5ab0eee71\"".to_string()),
 				size: IntValue(1111),
 				storage_class: Value("STANDARD".to_string()),
 			}],
@@ -618,7 +738,7 @@ mod tests {
   <Contents>\
     <Key>happyfacex.jpg</Key>\
     <LastModified>1970-01-01T00:00:00.000Z</LastModified>\
-    <ETag>70ee1738b6b21e2c8a43f3a5ab0eee71</ETag>\
+    <ETag>&quot;70ee1738b6b21e2c8a43f3a5ab0eee71&quot;</ETag>\
     <Size>1111</Size>\
     <StorageClass>STANDARD</StorageClass>\
   </Contents>\
@@ -626,4 +746,85 @@ mod tests {
 		);
 		Ok(())
 	}
+
+	#[test]
+	fn list_parts() -> Result<(), ApiError> {
+		let result = ListPartsResult {
+			xmlns: (),
+			bucket: Value("example-bucket".to_string()),
+			key: Value("example-object".to_string()),
+			upload_id: Value(
+				"XXBsb2FkIElEIGZvciBlbHZpbmcncyVcdS1tb3ZpZS5tMnRzEEEwbG9hZA".to_string(),
+			),
+			part_number_marker: Some(IntValue(1)),
+			next_part_number_marker: Some(IntValue(3)),
+			max_parts: IntValue(2),
+			is_truncated: Value("true".to_string()),
+			parts: vec![
+				PartItem {
+					etag: Value("\"7778aef83f66abc1fa1e8477f296d394\"".to_string()),
+					last_modified: Value("2010-11-10T20:48:34.000Z".to_string()),
+					part_number: IntValue(2),
+					size: IntValue(10485760),
+				},
+				PartItem {
+					etag: Value("\"aaaa18db4cc2f85cedef654fccc4a4x8\"".to_string()),
+					last_modified: Value("2010-11-10T20:48:33.000Z".to_string()),
+					part_number: IntValue(3),
+					size: IntValue(10485760),
+				},
+			],
+			initiator: Initiator {
+				display_name: Value("umat-user-11116a31-17b5-4fb7-9df5-b288870f11xx".to_string()),
+				id: Value(
+					"arn:aws:iam::111122223333:user/some-user-11116a31-17b5-4fb7-9df5-b288870f11xx"
+						.to_string(),
+				),
+			},
+			owner: Owner {
+				display_name: Value("someName".to_string()),
+				id: Value(
+					"75aa57f09aa0c8caeab4f8c24e99d10f8e7faeebf76c078efc7c6caea54ba06a".to_string(),
+				),
+			},
+			storage_class: Value("STANDARD".to_string()),
+		};
+
+		assert_eq!(
+			to_xml_with_header(&result)?,
+			"<?xml version=\"1.0\" encoding=\"UTF-8\"?>\
+<ListPartsResult xmlns=\"http://s3.amazonaws.com/doc/2006-03-01/\">\
+  <Bucket>example-bucket</Bucket>\
+  <Key>example-object</Key>\
+  <UploadId>XXBsb2FkIElEIGZvciBlbHZpbmcncyVcdS1tb3ZpZS5tMnRzEEEwbG9hZA</UploadId>\
+  <PartNumberMarker>1</PartNumberMarker>\
+  <NextPartNumberMarker>3</NextPartNumberMarker>\
+  <MaxParts>2</MaxParts>\
+  <IsTruncated>true</IsTruncated>\
+  <Part>\
+    <ETag>&quot;7778aef83f66abc1fa1e8477f296d394&quot;</ETag>\
+    <LastModified>2010-11-10T20:48:34.000Z</LastModified>\
+    <PartNumber>2</PartNumber>\
+    <Size>10485760</Size>\
+  </Part>\
+  <Part>\
+    <ETag>&quot;aaaa18db4cc2f85cedef654fccc4a4x8&quot;</ETag>\
+    <LastModified>2010-11-10T20:48:33.000Z</LastModified>\
+    <PartNumber>3</PartNumber>\
+    <Size>10485760</Size>\
+  </Part>\
+  <Initiator>\
+      <DisplayName>umat-user-11116a31-17b5-4fb7-9df5-b288870f11xx</DisplayName>\
+      <ID>arn:aws:iam::111122223333:user/some-user-11116a31-17b5-4fb7-9df5-b288870f11xx</ID>\
+  </Initiator>\
+  <Owner>\
+    <DisplayName>someName</DisplayName>\
+    <ID>75aa57f09aa0c8caeab4f8c24e99d10f8e7faeebf76c078efc7c6caea54ba06a</ID>\
+  </Owner>\
+  <StorageClass>STANDARD</StorageClass>\
+</ListPartsResult>"
+		);
+
+		Ok(())
+	}
 }
